"""Asset of type IP.
This module takes care of preparing an IP asset, either single address, a range with mask for both v4 and v6."""
import logging
import ipaddress
import click
from ostorlab.cli import console as cli_console
from ostorlab.assets import ipv4
from ostorlab.assets import ipv6
from ostorlab.cli.scan.run import run

logger = logging.getLogger(__name__)

console = cli_console.Console()

@run.run.command(name='ip')
@click.argument('ip', required=True)
@click.pass_context
def ip_cli(ctx: click.core.Context, ip: str) -> None:
    """Run scan for IP asset."""
    runtime = ctx.obj['runtime']
    try:
        ip_network = ipaddress.ip_network(ip, strict=False)
        if ip_network.version == 4:
            asset = ipv4.IPv4(host=ip_network.network_address.exploded, mask=ip_network.prefixlen)
        elif ip_network.version == 6:
            asset = ipv6.IPv6(host=ip_network.network_address.exploded, mask=ip_network.prefixlen)
        else:
            raise NotImplementedError()

        logger.debug('scanning asset IP %s', asset)
        runtime.scan(title=ctx.obj['title'], agent_group_definition=ctx.obj['agent_group_definition'], asset=asset)
    except ValueError as e:
        console.error(f'{e}')
        click.Abort()



"""Module responsible for installing an agent : Pulling the image from the ostorlab store."""
from typing import Dict, Optional, Generator
import click
import docker
import docker.errors

from ostorlab.cli import console as cli_console
from ostorlab.apis.runners import runner as base_runner
from ostorlab.apis.runners import public_runner, authenticated_runner
from ostorlab.apis import agent_details as agent_details_api
from ostorlab import configuration_manager
from ostorlab.cli.agent.install import install_progress


console = cli_console.Console()


def _image_name_from_key(agent_key: str) -> str:
    """Generate docker image name from an agent key.

    Args:
        agent_key: agent key in the form agent/organization/name.

    Returns:
        image_name: image name in the form : agent_organization_name.
    """
    return agent_key.replace('/', '_').lower()


def _parse_repository_tag(repo_name: str, tag: str = None) -> tuple:
    """Get repo name and tag"""
    parts = repo_name.rsplit('@', 1)
    if len(parts) == 2:
        return tuple(parts)
    parts = repo_name.rsplit(':', 1)
    if len(parts) == 2 and '/' not in parts[1]:
        return tuple(parts)
    return repo_name, tag


def _pull_logs(docker_client: docker.DockerClient,
               repository: str,
               tag: Optional[str] = None) -> Generator[Dict, None, None]:
    """Generate logs of the docker pull method."""
    repository, tag = _parse_repository_tag(repository, tag)
    pull_log = docker_client.api.pull(repository, tag=tag, stream=True, decode=True)
    for log in pull_log:
        yield log


def _get_image(docker_client: docker.DockerClient,
               repository: str,
               tag: str = None) -> docker.models.images.Image:
    """Gets a docker image."""
    repository, tag = _parse_repository_tag(repository, tag)
    name = f'{repository}:{tag}'

    return docker_client.images.get(name)


def _is_image_present(docker_client: docker.DockerClient, image_name: str) -> bool:
    """Check if a docker image exists."""
    try:
        docker_client.images.get(image_name)
        return True
    except docker.errors.ImageNotFound:
        return False


def get_agent_details(agent_key: str) -> Dict:
    """Sends an API request with the agent key, and retrieve the agent information.

    Args:
        agent_key: the agent key in the form : agent/org/name

    Returns:
        dictionary of the agent information like : name, dockerLocation..

    Raises:
        click Exit exception with satus code 2 when API response is invalid.
    """
    config_manager = configuration_manager.ConfigurationManager()

    if config_manager.is_authenticated:
        runner = authenticated_runner.AuthenticatedAPIRunner()
    else:
        runner = public_runner.PublicAPIRunner()

    try:
        response = runner.execute(agent_details_api.AgentDetailsAPIRequest(agent_key))
    except base_runner.ResponseError as e:
        console.error('Requested resource not found.')
        raise click.exceptions.Exit(2) from e


    if 'errors' in response:
        error_message = f"""\b The provided agent key : {agent_key} does not correspond to any agent.
        Please make sure you have the correct agent key.
        """
        console.error(error_message)
        raise click.exceptions.Exit(2)
    else:
        agent_details = response['data']['agent']
        return agent_details



def install(agent_key: str, version: str = '') -> None:
    """Install an agent : Fetch the docker file location of the agent corresponding to the agent_key,
    and pull the image from the registry.

    Args:
        agent_key: key of the agent in agent/org/agentName format.
        version: version of the docker image.

    Returns:
        None

    Raises:
        click Exit exception with satus code 2 when the docker image does not exist.
    """

    agent_details = get_agent_details(agent_key)

    agent_docker_location = agent_details['dockerLocation']
    image_name = _image_name_from_key(agent_details['key'])

    try:
        docker_client = docker.from_env()

        if _is_image_present(docker_client, image_name):
            console.info(f'{agent_key} already exist.')
        else:
            console.info('Pulling the image from the ostorlab store.')

            pull_logs_generator = _pull_logs(docker_client, agent_docker_location)

            progress_foo = install_progress.AgentInstallProgress()
            progress_foo.display(pull_logs_generator)

            agent_image = _get_image(docker_client=docker_client, repository=agent_docker_location, tag=version)
            agent_image.tag(repository=image_name, tag=version)

    except docker.errors.ImageNotFound as e:
        error_message = f'Image of the provided agent : {agent_key} was not found.'
        console.error(error_message)
        raise click.exceptions.Exit(2) from e



"""Handles all authenticated API calls and behind the scenes operations such as authentication, validation, etc.

    Typical usage example:

    authenticated_runner = AuthenticatedAPIRunner(username, password, token_duration)
    authenticated_runner.authenticate()
"""

import logging

from typing import Dict, Optional

import requests
import click

from ostorlab.apis.runners import runner
from ostorlab.apis.runners import login_runner
from ostorlab.apis import create_api_key
from ostorlab.apis import request as api_request
from ostorlab.cli import console as cli_console


AUTHENTICATED_GRAPHQL_ENDPOINT = 'https://api.ostorlab.co/apis/graphql'


logger = logging.getLogger(__name__)
console = cli_console.Console()


class AuthenticationError(runner.Error):
    """Authentication Error."""


class AuthenticatedAPIRunner(runner.APIRunner):
    """Handles all authenticated API calls and behind the scenes operations such as authentication,
       validation, etc.
    """

    def __init__(self,
                 username: str = None,
                 password: str = None,
                 token_duration: str = None,
                 proxy: str = None,
                 verify: bool = True
                 ):
        """Constructs all the necessary attributes for the object.

        Args:
            username: the username (email) used to login.
            password: the password used to login.
            token_duration: The duration for which the token is valid
            (Can be in minutes, hours, days, or a combination of any two or all three).
            proxy: The proxy through which a request is made. Defaults to None.
            verify: Whether or not to verify the TLS certificate. Defaults to True.
        """
        super().__init__(proxy, verify)
        self._username = username
        self._password = password
        self._token_duration = token_duration
        self._api_key: Optional[str] = self._configuration_manager.get_api_key()
        self._token: Optional[str] = None
        self._otp_token: Optional[str] = None


    @property
    def endpoint(self) -> str:
        """API endpoint."""
        return AUTHENTICATED_GRAPHQL_ENDPOINT


    def authenticate(self) -> None:
        """Authenticates the user.

        Raises:
            AuthenticationError: If user credentials are not valid.
        """
        with console.status('Logging into your account'):
            loggin_api_runner = login_runner.LoginAPIRunner(self._username,
                                                            self._password,
                                                            self._token,
                                                            self._proxy,
                                                            self._verify)
            response = loggin_api_runner.login_user()

        if response.status_code != 200:
            field_errors = response.json().get('non_field_errors')
            if field_errors is not None and field_errors[0] == 'Must include \"otp_token\"':
                self._otp_token = click.prompt(
                    'Please enter the OTP code from your authenticator app')
                self.authenticate()
            else:
                logger.debug('response %s', response.content)
                raise AuthenticationError(response.status_code)
        else:
            self._token = response.json().get('token')
            with console.status('Generating API key'):
                api_key_response = self.execute(
                    create_api_key.CreateAPIKeyAPIRequest(self._token_duration))
                console.success('API key generated')

            api_data = api_key_response['data']['createApiKey']['apiKey']
            secret_key = api_data['secretKey']
            api_key_id = api_data['apiKey']['id']
            expiry_date = api_data['apiKey']['expiryDate']

            self._api_key = secret_key
            with console.status('Persisting API key'):
                self._configuration_manager.set_api_data(
                    secret_key, api_key_id, expiry_date)
                console.success('API key persisted')
            self._token = None
            console.success(':white_check_mark: Authentication successful')

    def unauthenticate(self) -> None:
        self._api_key = None

    def execute(self, request: api_request.APIRequest) -> Dict:
        """Executes a request using the Authenticated GraphQL API.

        Args:
            request: The request to be executed

        Raises:
            ResponseError: When the API returns an error

        Returns:
            The API response
        """
        if self._token is not None:
            headers = {'Authorization': f'Token {self._token}'}
        elif self._api_key is not None:
            headers = {'X-Api-Key': f'{self._api_key}'}
        else:
            headers = None
            console.warning('No authentication credentials were provided.')

        response = self._sent_request(request, headers)
        if response.status_code != 200:
            raise runner.ResponseError(
                f'Response status code is {response.status_code}: {response.content}')
        else:
            return response.json()

    def _sent_request(self, request: api_request.APIRequest, headers=None,
                      multipart=False) -> requests.Response:
        """Sends an API request."""
        if self._proxy is not None:
            proxy = {
                'https': self._proxy
            }
        else:
            proxy = None
        if multipart:
            return requests.post(self.endpoint, files=request.data, headers=headers,
                                 proxies=proxy, verify=self._verify)
        else:
            return requests.post(self.endpoint, data=request.data, headers=headers,
                                 proxies=proxy, verify=self._verify)



"""IP address asset."""
import dataclasses
import ipaddress
from typing import Optional

from ostorlab.assets import asset


@dataclasses.dataclass
@asset.selector('v3.asset.ip')
class IP(asset.Asset):
    """IP Address target asset."""
    host: str
    version: Optional[int] = None
    mask: Optional[str] = None

    def __post_init__(self):
        """Sets the version to the correct value is not set at init."""
        if self.version is None:
            self.version = ipaddress.ip_address(self.host).version



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v2/source_code/file.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()

DESCRIPTOR = _descriptor.FileDescriptor(
    name='v2/source_code/file.proto',
    package='',
    serialized_pb=_b(
        '\n\x19v2/source_code/file.proto\"%\n\x04\x66ile\x12\x0f\n\x07scan_id\x18\x01 \x02(\x05\x12\x0c\n\x04\x66ile\x18\x02 \x02(\x0c')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_FILE = _descriptor.Descriptor(
    name='file',
    full_name='file',
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name='scan_id', full_name='file.scan_id', index=0,
            number=1, type=5, cpp_type=1, label=2,
            has_default_value=False, default_value=0,
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
        _descriptor.FieldDescriptor(
            name='file', full_name='file.file', index=1,
            number=2, type=12, cpp_type=9, label=2,
            has_default_value=False, default_value=_b(""),
            message_type=None, enum_type=None, containing_type=None,
            is_extension=False, extension_scope=None,
            options=None),
    ],
    extensions=[
    ],
    nested_types=[],
    enum_types=[
    ],
    options=None,
    is_extendable=False,
    extension_ranges=[],
    oneofs=[
    ],
    serialized_start=29,
    serialized_end=66,
)

DESCRIPTOR.message_types_by_name['file'] = _FILE

file = _reflection.GeneratedProtocolMessageType('file', (_message.Message,), dict(
    DESCRIPTOR=_FILE,
    __module__='v2.source_code.file_pb2'
    # @@protoc_insertion_point(class_scope:file)
))
_sym_db.RegisterMessage(file)

# @@protoc_insertion_point(module_scope)



"""Agent List command."""
import re
import logging
import click
import docker
from typing import Optional
from ostorlab.cli import console as cli_console
from ostorlab.cli.agent import agent

console = cli_console.Console()

logger = logging.getLogger(__name__)


@agent.command()
@click.option('--agent-key', '-a', help='Agent Key.', required=True)
@click.option('--agent-version-regex', '-r', help='Agent version matching regular expression.', required=False)
def delete(agent_key: str, agent_version_regex: Optional[str] = None) -> None:
    """CLI command to list installed agents."""
    docker_client = docker.from_env()
    images = docker_client.images.list()
    agent_container_name = agent_key.replace('/', '_')
    deleted = False
    for im in images:
        for t in im.tags:
            if t.split(':')[0] == agent_container_name:
                agent_container_version = t.split(':')[1]
                if agent_version_regex is None or re.match(agent_version_regex, agent_container_version):
                    console.info(f'deleting container container [bold red]{t}[/]')
                    docker_client.images.remove(t, force=True)
                    console.success(f'container image [bold red]{t}[/] deleted successfully')
                    deleted = True

    if deleted is False:
        console.error(f'No agent matching [bold white]{agent_key}[/] were found')
